# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE POS (API) -> Rattail data importing
"""

import decimal
import logging

from sqlalchemy import orm
from sqlalchemy.orm.exc import NoResultFound

from corepos.api import CoreWebAPI

from rattail import importing
from rattail.db import model
from rattail.gpc import GPC
from rattail.util import OrderedDict
from rattail.db.util import normalize_full_name
from rattail_corepos import importing as corepos_importing
from rattail_corepos.corepos.util import get_core_members


log = logging.getLogger(__name__)


class FromCOREPOSToRattail(importing.ToRattailHandler):
    """
    Import handler for data coming from a CORE POS API.
    """
    host_title = "CORE-POS (API)"

    def get_importers(self):
        importers = OrderedDict()
        importers['Customer'] = CustomerImporter
        importers['Person'] = PersonImporter
        importers['CustomerPerson'] = CustomerPersonImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Vendor'] = VendorImporter
        importers['Product'] = ProductImporter
        return importers


class FromCOREPOSAPI(importing.Importer):
    """
    Base class for all CORE POS API data importers.
    """

    def setup(self):
        super(FromCOREPOSAPI, self).setup()
        self.establish_api()

    def establish_api(self):
        url = self.config.require('corepos.api', 'url')
        self.api = CoreWebAPI(url)

    def get_core_members(self):
        return get_core_members(self.api, progress=self.progress)


class CustomerImporter(FromCOREPOSAPI, importing.model.CustomerImporter):
    """
    Importer for customer data from CORE POS API.
    """
    key = 'number'
    supported_fields = [
        'id',
        'number',
        'name',
    ]

    def get_host_objects(self):
        return self.get_core_members()

    def normalize_host_object(self, member):

        # figure out the "account holder" customer for the member
        customers = member['customers']
        account_holders = [customer for customer in customers
                           if customer['accountHolder']]
        if len(account_holders) > 1:
            log.warning("member %s has %s account holders in CORE: %s",
                        member['cardNo'], len(account_holders), member)
        elif not account_holders:
            raise NotImplementedError("TODO: how to handle member with no account holders?")
        customer = account_holders[0]

        return {
            'id': member['customerAccountID'],
            'number': member['cardNo'],
            'name': normalize_full_name(customer['firstName'],
                                        customer['lastName']),
        }


class PersonImporter(FromCOREPOSAPI, corepos_importing.model.PersonImporter):
    """
    Importer for person data from CORE POS API.
    """
    key = 'corepos_customer_id'
    supported_fields = [
        'corepos_customer_id',
        'first_name',
        'last_name',
        'display_name',
        'customer_uuid',
        'customer_person_ordinal',
    ]

    def setup(self):
        super(PersonImporter, self).setup()
        model = self.config.get_model()

        self.customers = self.cache_model(model.Customer, key='id')

    def get_host_objects(self):

        # first get all member data from CORE API
        members = self.get_core_members()
        normalized = []

        # then collect all the "person" records
        def normalize(member, i):
            normalized.extend(self.get_person_objects_for_member(member))

        self.progress_loop(normalize, members,
                           message="Collecting Person data from CORE")
        return normalized

    def get_person_objects_for_member(self, member):
        """
        Return a list of Person data objects for the given Member.  This
        logic is split out separately so that datasync can leverage it too.
        """
        people = []

        # make sure we put the account holder first in the list!
        customers = sorted(member['customers'],
                           key=lambda cust: 1 if cust['accountHolder'] else 0,
                           reverse=True)

        for i, customer in enumerate(customers, 1):
            person = dict(customer)
            person['customer_person_ordinal'] = i
            people.append(person)

        return people

    def get_customer(self, id):
        if hasattr(self, 'customers'):
            return self.customers.get(id)

        try:
            return self.session.query(model.Customer)\
                               .filter(model.Customer.id == id)\
                               .one()
        except NoResultFound:
            pass

    def normalize_host_object(self, person):

        customer = self.get_customer(person['customerAccountID'])
        if not customer:
            log.warning("Rattail customer not found for customerAccountID: %s",
                        person['customerAccountID'])
            return

        return {
            'corepos_customer_id': int(person['customerID']),
            'first_name': person['firstName'],
            'last_name': person['lastName'],
            'display_name': normalize_full_name(person['firstName'],
                                                person['lastName']),
            'customer_uuid': customer.uuid,
            'customer_person_ordinal': person['customer_person_ordinal'],
        }


class CustomerPersonImporter(FromCOREPOSAPI, importing.model.CustomerPersonImporter):
    """
    Importer for customer-person linkage data from CORE POS API.

    Note that we don't use this one in datasync, it's just for nightly
    double-check.
    """
    key = ('customer_uuid', 'person_uuid')
    supported_fields = [
        'customer_uuid',
        'person_uuid',
        'ordinal',
    ]

    def setup(self):
        super(CustomerPersonImporter, self).setup()
        model = self.config.get_model()

        self.customers = self.cache_model(model.Customer, key='id')

        query = self.session.query(model.Person)\
                            .join(model.CorePerson)\
                            .filter(model.CorePerson.corepos_customer_id != None)
        self.people = self.cache_model(model.Person, query=query,
                                       key='corepos_customer_id',
                                       query_options=[orm.joinedload(model.Person._corepos)])

    def get_host_objects(self):

        # first get all member data from CORE API
        members = self.get_core_members()
        normalized = []

        # then collect all customer/person combination records
        def normalize(member, i):
            # make sure we put the account holder first in the list!
            customers = sorted(member['customers'],
                               key=lambda cust: 1 if cust['accountHolder'] else 0,
                               reverse=True)
            for i, customer in enumerate(customers, 1):
                normalized.append({
                    'customer_account_id': member['customerAccountID'],
                    'person_customer_id': customer['customerID'],
                    'ordinal': i,
                })

        self.progress_loop(normalize, members,
                           message="Collecting CustomerPerson data from CORE")
        return normalized

    def get_customer(self, id):
        if hasattr(self, 'customers'):
            return self.customers.get(id)

        try:
            return self.session.query(model.Customer)\
                               .filter(model.Customer.id == id)\
                               .one()
        except NoResultFound:
            pass

    def get_person(self, corepos_customer_id):
        if hasattr(self, 'people'):
            return self.people.get(corepos_customer_id)

        model = self.config.get_model()
        try:
            return self.session.query(model.Person)\
                               .join(model.CorePerson)\
                               .filter(model.CorePerson.corepos_customer_id == corepos_customer_id)\
                               .one()
        except NoResultFound:
            pass

    def normalize_host_object(self, cp):

        customer = self.get_customer(cp['customer_account_id'])
        if not customer:
            log.warning("Rattail customer not found for customerAccountID: %s",
                        cp['customer_account_id'])
            return

        person = self.get_person(int(cp['person_customer_id']))
        if not person:
            log.warning("Rattail person not found for customerID: %s",
                        cp['person_customer_id'])
            return

        return {
            'customer_uuid': customer.uuid,
            'person_uuid': person.uuid,
            'ordinal': cp['ordinal'],
        }


class DepartmentImporter(FromCOREPOSAPI, importing.model.DepartmentImporter):
    """
    Importer for department data from CORE POS API.
    """
    key = 'number'
    supported_fields = [
        'number',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_departments()

    def normalize_host_object(self, department):
        return {
            'number': int(department['dept_no']),
            'name': department['dept_name'],
        }


class SubdepartmentImporter(FromCOREPOSAPI, importing.model.SubdepartmentImporter):
    """
    Importer for subdepartment data from CORE POS API.
    """
    key = 'number'
    supported_fields = [
        'number',
        'name',
        'department_number',
    ]

    def get_host_objects(self):
        return self.api.get_subdepartments()

    def normalize_host_object(self, subdepartment):
        department_number = None
        if 'dept_ID' in subdepartment:
            department_number = int(subdepartment['dept_ID'])

        return {
            'number': int(subdepartment['subdept_no']),
            'name': subdepartment['subdept_name'],
            'department_number': department_number,
        }


class VendorImporter(FromCOREPOSAPI, corepos_importing.model.VendorImporter):
    """
    Importer for vendor data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'name',
        'abbreviation',
        'special_discount',
        'phone_number',
        'fax_number',
        'email_address',
    ]

    def get_host_objects(self):
        return self.api.get_vendors()

    def normalize_host_object(self, vendor):
        return {
            'corepos_id': int(vendor['vendorID']),
            'name': vendor['vendorName'],
            'abbreviation': vendor['vendorAbbreviation'] or None,
            'special_discount': decimal.Decimal(vendor['discountRate']),
            'phone_number': vendor.get('phone') or None,
            'fax_number': vendor.get('fax') or None,
            'email_address': vendor.get('email') or None,
        }


class ProductImporter(FromCOREPOSAPI, importing.model.ProductImporter):
    """
    Importer for product data from CORE POS API.
    """
    key = 'item_id'
    supported_fields = [
        'item_id',
        'upc',
        'brand_name',
        'description',
        'size',
        'weighed',
        'department_number',
        'subdepartment_number',
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_type',
        'food_stampable',
        # 'tax1',
        # 'tax2',
    ]

    def get_host_objects(self):
        return self.api.get_products()

    def normalize_host_object(self, product):
        try:
            upc = GPC(product['upc'], calc_check_digit='upc')
        except (TypeError, ValueError):
            log.debug("CORE POS product has invalid UPC: %s", product['upc'])
            if len(self.key) == 1 and self.key[0] == 'upc':
                return
            upc = None

        department_number = None
        if 'department' in product:
            department_number = int(product['department'])

        subdepartment_number = None
        if 'subdept' in product:
            subdepartment_number = int(product['subdept']) or None

        price = None
        if product['normal_price'] is not None:
            price = decimal.Decimal(product['normal_price'])

        return {
            'item_id': product['upc'],
            'upc': upc,
            'brand_name': product.get('brand') or None,
            'description': product.get('description') or '',
            'size': product.get('size', '').strip() or None,

            'department_number': department_number,
            'subdepartment_number': subdepartment_number,

            'weighed': product['scale'] == '1',
            'food_stampable': product['foodstamp'] == '1',
            # 'tax1': product['tax'] == '1', # TODO: is this right?
            # 'tax2': product['tax'] == '2', # TODO: is this right?

            'regular_price_price': price,
            'regular_price_multiple': 1 if price is not None else None,
            'regular_price_type': self.enum.PRICE_TYPE_REGULAR if price is not None else None,
        }
